import argparse
import pathlib
import sys
from typing import List, Union

import tekstialue.tekstialue as api
from tekstialue import APP_ALIAS, APP_NAME, DEFAULT_CONFIG_NAME, log


def parse_request(argv: List[str]) -> Union[int, argparse.Namespace]:
    """DRY."""
    parser = argparse.ArgumentParser(
        prog=APP_ALIAS, description=APP_NAME, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        '--config',
        '-c',
        dest='cfg_file',
        default='',
        help=f'Configuration file to read column defs from. Optional\n(default: {DEFAULT_CONFIG_NAME})',
        required=False,
    )
    parser.add_argument(
        '--input',
        '-i',
        dest='in_file',
        default='',
        help='File to transform from. Optional\n(default: positional in-file (first) value)',
        required=False,
    )
    parser.add_argument(
        'in_file_pos',
        nargs='?',
        default='',
        help='File to transform from. Optional\n(default: value of input option)',
    )
    parser.add_argument(
        '--output',
        '-o',
        dest='out_file',
        default='',
        help='File to transform to. Optional\n(default: in-file.out)',
        required=False,
    )
    parser.add_argument(
        '--from',
        '-f',
        dest='from_format',
        default='latex',
        help='format to transform from (default: latex)',
        required=False,
    )
    parser.add_argument(
        '--to',
        '-t',
        dest='to_format',
        default='latex',
        help='format to transform to (default: latex)',
        required=False,
    )
    parser.add_argument(
        '--verbose',
        '-v',
        dest='verbose',
        default=False,
        action='store_true',
        help='work logging more information along the way (default: False)',
    )
    parser.add_argument(
        '--debug',
        '-d',
        dest='debug',
        default=False,
        action='store_true',
        help='be even more vrbose to support debugging (default: False)',
    )
    if not argv:
        parser.print_help()
        return 0

    options = parser.parse_args(argv)

    if not options.in_file:
        if options.in_file_pos:
            options.in_file = options.in_file_pos

    if not options.out_file:
        options.out_file = f'{options.in_file}.out'

    if options.from_format != 'latex':
        log.error(f'unsupported from format {options.from_format} - only latex currently supported')
        return 2

    if options.to_format != 'latex':
        log.error(f'unsupported to format {options.to_format} - only latex currently supported')
        return 2

    if pathlib.Path(options.in_file).is_file():
        return options

    log.error(f'input {options.in_file} does not exist or is no file')
    return 1


def main(argv: Union[List[str], None] = None) -> int:
    """Delegate processing to functional module."""
    argv = sys.argv[1:] if argv is None else argv
    options = parse_request(argv)
    if isinstance(options, int):
        return options
    return api.main(options)  # type: ignore



import pathlib

import tekstialue.tekstialue as api

TEST_PREFIX = pathlib.Path('test', 'fixtures', 'basic')
DEFAULT_CFG_PATH = TEST_PREFIX / 'dot.tekstialue.json'
DEFAULT_REAL_IN_FILE_PATH = TEST_PREFIX / 'real-la.tex.in'


def test_cue_tables():
    assert api.cue_tables([]) == []


def test_cue_tables_for_real():
    with open(DEFAULT_REAL_IN_FILE_PATH, 'rt', encoding=api.ENCODING) as handle:
        lines = [''] + [line.rstrip() for line in handle.readlines()]
    assert api.cue_tables(lines) == [
        {
            'start': 2,
            'end_data_row': [21, 24],
            'top_rule': 7,
            'mid_rule': 17,
            'end_head': 18,
            'bottom_rule': 25,
            'end': 26,
            'amend': 28,
        }
    ]


def test_discover_configuration():
    code, cfg, cp = api.discover_configuration(DEFAULT_CFG_PATH)
    assert not code
    assert cp == str(DEFAULT_CFG_PATH)
    cols = cfg['columns']
    assert cols['col_1'] == ['a']
    assert cols['col_2'] == ['b']
    assert cols['col_3'] == ['c', 'cc']
    assert cols['col_4'] == ['d', 'dd']


def test_extract_slots():
    assert api.extract_slots([{'start': 0, 'amend': 42}]) == [(0, 42 + 1)]


def test_weave_table():
    assert api.weave_table([''], [], [], '') == ['']



import pathlib

import tekstialue.cli as cli

TEST_PREFIX = pathlib.Path('test', 'fixtures', 'basic')
DEFAULT_IN_FILE_PATH = TEST_PREFIX / 'in-file.tex'
DEFAULT_CFG_PATH = TEST_PREFIX / 'dot.tekstialue.json'
TEST_MAKE_MISSING = 'missing-this-file-for-'


def test_parse_request(capsys):
    options = cli.parse_request(['-i', str(DEFAULT_IN_FILE_PATH)])
    assert options.in_file_pos == ''  # type: ignore
    assert options.in_file == str(DEFAULT_IN_FILE_PATH)  # type: ignore
    out, err = capsys.readouterr()
    assert not out
    assert not err


def test_parse_request_verbose(capsys):
    options = cli.parse_request(['-i', str(DEFAULT_IN_FILE_PATH), '-v'])
    assert options.verbose
    out, err = capsys.readouterr()
    assert not out
    assert not err


def test_parse_request_config(capsys):
    options = cli.parse_request(['-i', str(DEFAULT_IN_FILE_PATH), '-c', str(DEFAULT_CFG_PATH)])
    assert options.cfg_file == str(DEFAULT_CFG_PATH)
    out, err = capsys.readouterr()
    assert not out
    assert not err


def test_parse_request_config_home(capsys):
    options = cli.parse_request(['-i', str(DEFAULT_IN_FILE_PATH)])
    assert options.cfg_file == ''
    out, err = capsys.readouterr()
    assert not out
    assert not err


def test_parse_empty_request(capsys):
    options = cli.parse_request([])
    assert options == 0
    out, err = capsys.readouterr()
    assert 'input' in out
    assert not err



"""Purge monotonically named files in folders keeping range endpoints.

Implementation uses sha256 hashes for identity and assumes that
the natural order relates to the notion of fresher or better.
"""
import argparse
import datetime as dti
import json
import logging
import pathlib
from typing import no_type_check

from tekstialue import DEFAULT_CONFIG_NAME, ENCODING, log

TAB_START_TOK = r'\begin{longtable}[]{@{}'
TOP_RULE = r'\toprule()'
MID_RULE = r'\midrule()'
END_HEAD = r'\endhead'
END_DATA_ROW = r'\\'
BOT_RULE = r'\bottomrule()'
TAB_END_TOK = r'\end{longtable}'

TAB_NEW_START = r"""\begin{small}
\begin{longtable}[]{|
>{\raggedright\arraybackslash}p{(\columnwidth - 12\tabcolsep) * \real{0.1500}}|
>{\raggedright\arraybackslash}p{(\columnwidth - 12\tabcolsep) * \real{0.5500}}|
>{\raggedright\arraybackslash}p{(\columnwidth - 12\tabcolsep) * \real{0.1500}}|
>{\raggedright\arraybackslash}p{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|}
\hline"""

TAB_HACKED_HEAD = r"""\begin{minipage}[b]{\linewidth}\raggedright
\ \mbox{\textbf{$COL1$}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{$COL2$}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{$COL3_A$}} \mbox{\textbf{$COL3_B$}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{$COL4_A$}} \mbox{\textbf{$COL4_B$}}
\end{minipage} \\
\hline
\endfirsthead
\multicolumn{4}{@{}l}{\small \ldots continued}\\\hline
\hline
\begin{minipage}[b]{\linewidth}\raggedright
\ \mbox{\textbf{$COL1$}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{$COL2$}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{$COL3_A$}} \mbox{\textbf{$COL3_B$}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{$COL4_A$}} \mbox{\textbf{$COL4_B$}}
\end{minipage} \\
\endhead
\hline"""

COL_1 = 'a'
COL_2 = 'b'
COL_3_A = 'c'
COL_3_B = 'cc'
COL_4_A = 'd'
COL_4_B = 'dd'

NEW_RULE = r'\hline'

TAB_NEW_END = r"""\end{longtable}
\end{small}
\vspace*{-2em}
\begin{footnotesize}
ANNOTATION
\end{footnotesize}"""

Slots = list[tuple[int, int]]
TableRanges = list[dict[str, int | list[int]]]


@no_type_check
def discover_configuration(conf: str) -> tuple[int, dict[str, object], str]:
    """Try to retrieve the configuration following the "(explicit, local, parents, home)
    first wun wins" strategy."""
    configuration = None
    if conf:
        cp = pathlib.Path(conf)
        if not cp.is_file() or not cp.stat().st_size:
            log.error('Given configuration path is no file or empty')
            return 1, {}, ''
        log.debug(f'Reading configuration file {cp} as requested...')
        with cp.open(encoding=ENCODING) as handle:
            configuration = json.load(handle)
    else:
        cn = DEFAULT_CONFIG_NAME
        cwd = pathlib.Path.cwd().resolve()
        for pp in (cwd, *cwd.parents):
            cp = pp / cn
            if cp.is_file() and cp.stat().st_size:
                log.debug(f'Reading from discovered configuration path {cp}')
                with cp.open() as handle:
                    configuration = json.load(handle)
                return 0, configuration, str(cp)

        cp = pathlib.Path.home() / DEFAULT_CONFIG_NAME
        if cp.is_file() and cp.stat().st_size:
            log.debug(f'Reading configuration file {cp} from home directory at {pathlib.Path.home()} ...')
            with cp.open() as handle:
                configuration = json.load(handle)
            return 0, configuration, str(cp)

        log.debug(f'User home configuration path to {cp} is no file or empty - ignoring configuration data')

    return 0, configuration, str(cp)


@no_type_check
def cue_tables(lines: list[str]) -> TableRanges:
    """Tag all tables extracting the relevant line information for elements."""
    table_section, head, annotation = False, False, False
    table_ranges = []
    guess_slot = 0
    table_range = {}
    for n, text in enumerate(lines):
        if not table_section:
            if not text.startswith(TAB_START_TOK):
                continue
            table_range['start'] = n
            table_section = True
            head = True
            table_range['end_data_row'] = []
            continue

        if text.startswith(TOP_RULE):
            table_range['top_rule'] = n
            continue

        if text.startswith(MID_RULE):
            table_range['mid_rule'] = n
            continue

        if text.startswith(END_HEAD):
            table_range['end_head'] = n
            head = False
            continue

        if not head and text.strip().endswith(END_DATA_ROW):
            table_range['end_data_row'].append(n)
            continue

        if text.startswith(BOT_RULE):
            table_range['bottom_rule'] = n
            continue

        if text.startswith(TAB_END_TOK):
            table_range['end'] = n
            annotation = True
            guess_slot = n + 2
            continue

        if annotation and n == guess_slot:
            table_range['amend'] = n
            table_ranges.append(table_range)
            table_range = {}
            annotation, table_section = False, False

    return table_ranges


@no_type_check
def extract_slots(table_ranges: TableRanges) -> Slots:
    """Extract the on and off slots for output processing."""
    on_off_slots = []
    for table in table_ranges:
        from_here = table['start']
        thru_there = table['amend']
        on_off = (from_here, thru_there + 1)
        on_off_slots.append(on_off)

    return on_off_slots


@no_type_check
def weave_table(lines: list[str], on_off_slots: Slots, table_ranges: TableRanges, tab_hacked_head: str) -> list[str]:
    """Generate the output."""
    out = []
    next_slot = 0
    for n, line in enumerate(lines):
        if next_slot < len(on_off_slots):
            trigger_on, trigger_off = on_off_slots[next_slot]
            tb = table_ranges[next_slot]
        else:
            trigger_on = None
        if trigger_on is None:
            out.append(line)
            continue

        if n < trigger_on:
            out.append(line)
            continue
        if n == trigger_on:
            out.append(TAB_NEW_START)
            out.append(tab_hacked_head)
            continue
        if n <= tb['end_head']:
            continue
        if n < tb['bottom_rule']:
            out.append(line)
            if n in tb['end_data_row']:
                out.append(NEW_RULE)
            continue
        if tb['bottom_rule'] <= n < tb['amend']:
            continue
        if n == tb['amend']:
            out.append(TAB_NEW_END.replace('ANNOTATION', line))
            next_slot += 1

    return out


@no_type_check
def main(options: argparse.Namespace) -> int:
    """Process the text."""
    start_time = dti.datetime.utcnow()
    verbose = options.verbose
    if verbose:
        logging.getLogger().setLevel(logging.DEBUG)
    in_file, out_file = options.in_file, options.out_file

    code, cfg, cp = discover_configuration(options.cfg_file)
    if code:
        return code
    log.info(f'Read configiration from {cp}')
    log.debug(f'{cfg=}')
    cols = cfg['columns']
    tab_hacked_head = TAB_HACKED_HEAD.replace('$COL1$', cols['col_1'][0])
    tab_hacked_head = tab_hacked_head.replace('$COL2$', cols['col_2'][0])
    tab_hacked_head = tab_hacked_head.replace('$COL3_A$', cols['col_3'][0])
    tab_hacked_head = tab_hacked_head.replace('$COL3_B$', cols['col_3'][1])
    tab_hacked_head = tab_hacked_head.replace('$COL4_A$', cols['col_4'][0])
    tab_hacked_head = tab_hacked_head.replace('$COL4_B$', cols['col_4'][1])

    with open(in_file, 'rt', encoding=ENCODING) as handle:
        lines = [''] + [line.rstrip() for line in handle.readlines()]
    log.debug(f'Read {len(lines)} lines from {in_file}')

    table_ranges = cue_tables(lines)
    on_off_slots = extract_slots(table_ranges)

    out = weave_table(lines, on_off_slots, table_ranges, tab_hacked_head)
    with open(out_file, 'wt', encoding=ENCODING) as handle:
        handle.write('\n'.join(out) + '\n')
    log.debug(f'Wrote {len(lines)} lines to {out_file}')

    duration_seconds = (dti.datetime.utcnow() - start_time).total_seconds()

    log.info(f'transformed tables in {in_file} into {out_file}' f' in {duration_seconds} secs')
    return 0
